#ifndef FAST_LRU_H
#define FAST_LRU_H
/*
 * Filename      : lru.h
 * Purpose       : This program provides an efficient implementation
 *                 of LRU list using hash table and buckets.
 * Author        : Lei Yang
 * Date Created  : 11/27/2006
 * Last Modified : 11/288888888/2006
*/
#define USE_HASH_MAP 1
//#define PRINT_BUCKETS 1
#define BUCKET_SIZE_BASE 4 

#include <list>
#include <map>
#include <iostream>
#include "assert.h"
#include "math.h"

#ifdef USE_HASH_MAP
#include <ext/hash_map>
#endif

using namespace std;

/****************************************************************************/
template<class T>
class ListNode
{
public:
	ListNode(T index, int num)
	{
		id = index;
		bucket_num = num;
	}
  
  ListNode()
  {
  }
  
  bool operator== (const ListNode &node)
  {
    if ((node.id == id) && (node.bucket_num == bucket_num))
      return true;
    else
      return false;
  }
  
  bool operator!= (const ListNode &node)
  {
    return !(*this == node);
  }
  
	T id;
	int bucket_num;
};

template <class T>
ostream &operator<< (ostream &output, const ListNode<T> & node)
{
  output << "<" << node.id << "," << node.bucket_num << "> ";
  return output;
}

/****************************************************************************/
template<class T>
class lru
{
public:	
	lru()
	{
	}
	
	~lru()
	{
		mymap.clear();
		buckets.clear();
	}
	
	void insert(const T &id)
  {
    assert(find(id) == -1);
		node_type newnode(id, 0);
		
    // This is the very first node, create a new bucket,
    // push this node to the front of the new bucket, and
    // add the new bucket to the end of the bucket list
    if (buckets.empty())
    {
      bucket_type newbucket;
      newbucket.push_front(newnode);
      buckets.push_front(newbucket);
      mymap[id] = (buckets.begin())->begin();
    }
    // Push to the front of the first bucket
    else
    {
      bucket_iter bi = buckets.begin();
      bi->push_front(newnode);
      mymap[id] = bi->begin();
	
      // Push back node to the next bucket if this bucket is full
      for (int i = 0; ; i++)
      {
        if (bi->size() > pow(BUCKET_SIZE_BASE,(i+1)))
        {
          node_type tmp = bi->back();
          tmp.bucket_num = i + 1;
          bi->pop_back();
          
          bi++;
          if (bi == buckets.end()) 
          {
            bucket_type newbucket;
            newbucket.push_front(tmp);
            buckets.push_back(newbucket);
            mymap[tmp.id] = (buckets.rbegin())->begin();
            break;
          }
          else
          {
            bi->push_front(tmp);
            mymap[tmp.id] = bi->begin();
          }
        }
        else
        {
          break;
        }
      }
    }
#ifdef PRINT_BUCKETS 
    print_buckets();
#endif
  }

	int find(const T &id) 
  {
		int num = -1;
    map_iter mi = mymap.find(id);
    
		if (mi == mymap.end())
		{
      return num;
    }
		else
		{
			node_iter ni = mi->second;
      bucket_iter bi = buckets.begin();
      num = ni->bucket_num;
      
      // Not in the first bucket, move to first, 
      // then update hash map
			if (num != 0)
			{
				node_type tmp(ni->id, 0);
				
				bi->push_front(tmp);
				mymap[tmp.id] = bi->begin();
        
        advance(bi, num);
				bi->erase(ni);
        advance(bi, -num);
        
        // Push back node to the next bucket if this bucket is full
        for (int i = 0; bi != buckets.end(); i++)
        {
          if (bi->size() > pow(BUCKET_SIZE_BASE, (i+1)))
          {
            node_type tmp = bi->back();
            tmp.bucket_num = i + 1;
            bi->pop_back();
          
            bi++;
            bi->push_front(tmp);
			
            mymap[tmp.id] = bi->begin();
          }
          else
          {
            break;
          }
        }
        return num;
			}
      // In the first bucket, but not in the front, move to front,
      // then update hash map
			else if (*ni != bi->front())
			{
				node_type tmp = *ni ;
				bi->erase(ni);
				bi->push_front(tmp);
				mymap[tmp.id] = bi->begin();
        return num;
      }
      // In the first bucket, also in the front, nothing needs be done
      else
      {
        return num;
      }
		}
	}

	const int size()
	{
		return mymap.size();
	}
  
  void print_buckets()
  {
    bucket_iter bi = buckets.begin();
    
    for (int i = 0; bi != buckets.end(); bi++, i++)
    {
      cout << "bucket" << i << ": ";
      if (bi->empty())
      {
        cout << "empty" << endl;
        continue;
      }
      node_iter ni;
      for (ni = bi->begin(); ni != bi->end(); ni++)
      {
        cout << (*ni);
      }
      cout << endl;
    }
  }
  
private:
  typedef ListNode<T> node_type;
  typedef list<node_type> bucket_type;
  typedef list<bucket_type> bucket_list_type;
  
  typedef typename bucket_type::iterator node_iter;
  typedef typename bucket_list_type::iterator bucket_iter;
  
  #ifdef USE_HASH_MAP
  typedef __gnu_cxx::hash_map<T, node_iter> map_type;
  #else
  typedef std::map<T, node_iter> map_type;
  #endif
  
  typedef typename map_type::iterator map_iter;
   
  map_type mymap;
  bucket_list_type buckets;
};

#endif
